/**
 *
 *  You can modify and use this source freely
 *  only for the development of application related Live2D.
 *
 *  (c) Live2D Inc. All rights reserved.
 */

#import "ViewController.h"
#import "Live2DModelIPhone.h"
#import "util/UtSystem.h"
#import "motion/MotionQueueManager.h"
#import "motion/Live2DMotion.h"

#import <vector>
#import <sstream>

using namespace live2d;

const int NUM_MODELS = 5;

@interface ViewController ()

@property (strong, nonatomic) EAGLContext *context;

- (void)setupGL;
- (void)tearDownGL;

@end

@implementation ViewController
{
	std::vector<Live2DModelIPhone*> models;
	std::vector<GLuint> textures ;
	MotionQueueManager motionMgr;
	Live2DMotion* motion;
    UILabel* label;
	
	CFTimeInterval lastTime;
	int fps;
}

- (void)dealloc
{
	// Live2Dモデルの解放
	for (int i = 0; i < NUM_MODELS; i++)
	{
		delete models[i];
	}
	
	// テクスチャを解放
	for (int i=0; i<textures.size(); i++)
	{
		GLuint glTexNo=textures[i];
		glDeleteTextures(1,&glTexNo);
	}

	delete motion;
	
    [self tearDownGL];
    
    if ([EAGLContext currentContext] == self.context) {
        [EAGLContext setCurrentContext:nil];
    }
    
    [_context release];
    [super dealloc];
}

- (void)viewDidLoad
{
    [super viewDidLoad];
    
    self.context = [[[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES1] autorelease];

    if (!self.context) {
        NSLog(@"Failed to create ES context");
    }
    
	lastTime= CFAbsoluteTimeGetCurrent();    // 現在の時刻を取得
    fps = 0;
	
    GLKView *view = (GLKView *)self.view;
    view.context = self.context;
    view.drawableDepthFormat = GLKViewDrawableDepthFormat24;
    
    [self setupGL];
    
    label = [[UILabel alloc] initWithFrame:CGRectMake(10, 10, 320, 50)];
    label.text = @"FPS : 0";
    [self.view addSubview:label];
    
    self.preferredFramesPerSecond = 60;
	
	NSString* MODEL_FILE = @"haru" ;
	NSString* TEXTURES[] ={
		@"texture_00" ,
		@"texture_01" ,
		@"texture_02" ,
		NULL
	} ;
	NSString* MOTION_FILE = @"idle_00" ;
	
	for( int i = 0 ; TEXTURES[i] != NULL ; i++ ){
		NSString* texturePath = [[NSBundle mainBundle] pathForResource:TEXTURES[i] ofType:@"png"];
		GLKTextureInfo* textureInfo =
		[GLKTextureLoader textureWithContentsOfFile:texturePath
											options:[NSDictionary dictionaryWithObjectsAndKeys:
													 [NSNumber numberWithBool:YES] ,GLKTextureLoaderApplyPremultiplication,
													 [NSNumber numberWithBool:YES] ,GLKTextureLoaderGenerateMipmaps,nil
													 ]
											  error:nil];
		
		int glTexNo = [textureInfo name] ;
		textures.push_back(glTexNo);// 解放するために設定
	}
	
		
	for (int i = 0; i < NUM_MODELS; i++)
	{
		NSString *modelpath = [[NSBundle mainBundle] pathForResource:MODEL_FILE ofType:@"moc"];
		
		Live2DModelIPhone* live2DModel = Live2DModelIPhone::loadModel( [modelpath UTF8String] ) ;
		live2DModel->setPremultipliedAlpha(true);
        
		models.push_back(live2DModel);
		
		for( int i = 0 ; TEXTURES[i] != NULL ; i++ ){
			live2DModel->setTexture( i , textures[i] ) ;// テクスチャとモデルを結びつける
		}
		
	}
	
	float modelWidth = models[0]->getCanvasWidth(); // モデルの幅
    float width = [[UIScreen mainScreen] bounds].size.width;
	float height = [[UIScreen mainScreen] bounds].size.height;
	//  描画範囲の設定
	glMatrixMode( GL_PROJECTION ) ;
	glLoadIdentity() ;
	glOrthof(
             0,
             modelWidth,
             modelWidth / (width/height),
             0,
             0.5f, -0.5f
             );
	
	NSString *motionpath = [[NSBundle mainBundle] pathForResource:MOTION_FILE ofType:@"mtn"];
	
	motion = Live2DMotion::loadMotion([motionpath UTF8String] ) ;
}

- (void)didReceiveMemoryWarning
{
    [super didReceiveMemoryWarning];

    if ([self isViewLoaded] && ([[self view] window] == nil)) {
        self.view = nil;
        
        [self tearDownGL];
        
        if ([EAGLContext currentContext] == self.context) {
            [EAGLContext setCurrentContext:nil];
        }
        self.context = nil;
    }

    // Dispose of any resources that can be recreated.
}

- (void)setupGL
{
    [EAGLContext setCurrentContext:self.context];
}

- (void)tearDownGL
{
    [EAGLContext setCurrentContext:self.context];
}

#pragma mark - GLKView and GLKViewController delegate methods

- (void)glkView:(GLKView *)view drawInRect:(CGRect)rect
{
	CFTimeInterval time = CFAbsoluteTimeGetCurrent();
	
    fps++;
    if ((time - lastTime) >= 1) {
        NSLog(@"FPS = %d", fps);
        
        std::ostringstream oss;
        oss << "FPS : " << fps;
        label.text = [[[NSString alloc] initWithCString:oss.str().c_str() encoding:NSUTF8StringEncoding] autorelease];
        
        fps = 0;
        lastTime = time;
    }
	
	glMatrixMode(GL_MODELVIEW ) ;
	glLoadIdentity() ;
    glClearColor(0.65f, 0.65f, 0.65f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	if (motionMgr.isFinished())
	{
		motionMgr.startMotion(motion,false);
	}
	
	float d=models[0]->getCanvasWidth()/NUM_MODELS/2.0f;
	

	for (int i = 0; i < NUM_MODELS; i++)
	{
		Live2DModelIPhone* live2DModel=models[i];
		motionMgr.updateParam(live2DModel);
		glTranslatef(d, 0, 0);
		live2DModel->update() ;
		live2DModel->draw() ;
	}
}

@end
