/**
 *  EAGLView.mm
 *  Live2DSample
 *
 *  Copyright 2014 Live2D Inc. All rights reserved.
 */

#import "EAGLView.h"


#import "Live2D.h"
#import "UtSystem.h"

using namespace live2d ;

@implementation EAGLView

NSString* MODEL_PATH = @"haru" ;
NSString* TEXTURE_PATH[] = {
    @"texture_00.png" ,
    @"texture_01.png" ,
    @"texture_02.png" ,
    NULL
} ;

//==================================================
// You must implement this method
//==================================================
+ (Class) layerClass
{
    return [CAEAGLLayer class];
}

//==================================================
//The GL view is stored in the nib file. When it's unarchived it's sent -initWithCoder:
//==================================================
- (id)initWithFrame:(CGRect)frame {
    if ((self = [super initWithFrame:frame])) {
		
        // Get the layer
        CAEAGLLayer *eaglLayer = (CAEAGLLayer *)self.layer;
        self.contentScaleFactor = [UIScreen mainScreen].scale ;
        eaglLayer.opaque = TRUE;		
        
		context = [[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES1];
        
        if (!context || ![EAGLContext setCurrentContext:context])
		{
            [self release];
            return nil;
        }
		glGenFramebuffersOES(1, &defaultFramebuffer);
		glGenRenderbuffersOES(1, &colorRenderbuffer);
		glBindFramebufferOES(GL_FRAMEBUFFER_OES, defaultFramebuffer);
		glBindRenderbufferOES(GL_RENDERBUFFER_OES, colorRenderbuffer);
		glFramebufferRenderbufferOES(GL_FRAMEBUFFER_OES, GL_COLOR_ATTACHMENT0_OES, GL_RENDERBUFFER_OES, colorRenderbuffer);
        
		
		// 以下の命令でメモリリークの検査を実施（_DEBUGモードのみ）
		// 　Live2D::dispose()時に、Live2Dの管理するメモリでリークが発生していた場合にダンプする
		// 	リークしている場合は、MEMORY_DEBUG_MEMORY_INFO_ALLでより詳細な情報をダンプします
		// 　検査用のデータはglobal new演算子を使います
		// live2d::UtDebug::addMemoryDebugFlags( live2d::UtDebug::MEMORY_DEBUG_MEMORY_INFO_COUNT ) ;//メモリリークの検出用
		
        Live2D::init() ;
		
		
	

        //  モデルの設定
        NSString *modelpath = [[NSBundle mainBundle] pathForResource:MODEL_PATH ofType:@"moc"];
        live2DModel = Live2DModelIPhone::loadModel( [modelpath UTF8String] ) ;
        
		textures = [[NSMutableArray array] retain];
        for( int i = 0 ; TEXTURE_PATH[i] != NULL ; i++ )
        {
            int texNo = [self loadTexture:(TEXTURE_PATH[i])] ;
            live2DModel->setTexture( i , texNo ) ;// テクスチャとモデルを結びつける
			[textures addObject:[NSNumber numberWithInt:texNo]];// 解放するために設定
        }
		//Live2D::dispose();

    }
    return self;
}

//==================================================
//	DRAW
//==================================================
- (void) drawView:(id)sender
{
	[EAGLContext setCurrentContext:context];
    
    //  モデル用のOpenGLの描画関係を設定
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glClear(GL_COLOR_BUFFER_BIT);
    glEnable(GL_BLEND);
	glBlendFunc(GL_ONE , GL_ONE_MINUS_SRC_ALPHA );
	glDisable(GL_DEPTH_TEST) ;
	glDisable(GL_CULL_FACE) ;
    
    // モデルのパラメータを変更。動作確認用です。
	double t = (UtSystem::getUserTimeMSec()/1000.0) * 2 * M_PI  ;// 1秒ごとに2π(1周期)増える
	double cycle=3.0;// パラメータが一周する時間(秒)
	double value=sin( t/cycle );// -1から1の間を周期ごとに変化する
	live2DModel->setParamFloat( "PARAM_ANGLE_X" , (float) (30 * value) ) ;// PARAM_ANGLE_Xのパラメータが[cycle]秒ごとに-30から30まで変化する
	

    
	live2DModel->update();
	live2DModel->draw();
	
    [context presentRenderbuffer:GL_RENDERBUFFER_OES];
}


//---------------------------------------------------
- (void) layoutSubviews
{
	[context renderbufferStorage:GL_RENDERBUFFER_OES fromDrawable:(CAEAGLLayer*)self.layer];
	glGetRenderbufferParameterivOES(GL_RENDERBUFFER_OES, GL_RENDERBUFFER_WIDTH_OES, &deviceWidth);
    glGetRenderbufferParameterivOES(GL_RENDERBUFFER_OES, GL_RENDERBUFFER_HEIGHT_OES, &deviceHeight);
	
    
	//  ビューポートはデバイスの幅と合わせる。画面全体に表示される。
	glViewport(0, 0, deviceWidth, deviceHeight);
	
    //  簡易的にプロジェクション行列一つですべての変換を行う。
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	
    float modelWidth = live2DModel->getCanvasWidth(); //  モデラーで設定したキャンバス幅
    
	//  描画範囲の設定。引数はleft, right, bottom, top の順。
	glOrthof(
             0,
             modelWidth,
             modelWidth * deviceHeight / deviceWidth,
             0,
             0.5f, -0.5f 
             );
	
    [NSTimer scheduledTimerWithTimeInterval:(NSTimeInterval)(1/60) 
                                     target:self 
                                   selector:@selector(drawView:) 
                                   userInfo:nil repeats:TRUE];
}


- (void) dealloc
{
	delete live2DModel;
	
	// テクスチャを解放
	for (int i=0; i<[textures count]; i++)
	{
		GLuint texNo=[[textures objectAtIndex:i] intValue];
		glDeleteTextures(1,&texNo);
	}
	[textures removeAllObjects];
	[textures release];
	
	Live2D::dispose();
	
	// Tear down GL
	if (defaultFramebuffer)
	{
		glDeleteFramebuffersOES(1, &defaultFramebuffer);
		defaultFramebuffer = 0;
	}
	
	if (colorRenderbuffer)
	{
		glDeleteRenderbuffersOES(1, &colorRenderbuffer);
		colorRenderbuffer = 0;
	}
	
	// Tear down context
	if ([EAGLContext currentContext] == context)
        [EAGLContext setCurrentContext:nil];
	
	[context release];
	context = nil;
    
	[super dealloc];
}


// 画像ファイルからテクスチャを作成
- (GLuint)loadTexture:(NSString*)fileName
{
	GLuint texture;
	
	// 画像ファイルを展開しCGImageRefを生成
	UIImage* uiImage = [UIImage imageNamed:fileName];
	
	CGImageRef image = uiImage.CGImage ;
	
	// 画像の大きさを取得
	size_t width = CGImageGetWidth(image);
	size_t height = CGImageGetHeight(image);
	
	// ビットマップデータを用意
	GLubyte* imageData = (GLubyte *) calloc(width * height * 4 , 1);
	CGContextRef imageContext = CGBitmapContextCreate(imageData,width,height,8,width * 4,CGImageGetColorSpace(image),kCGImageAlphaPremultipliedLast);
	CGContextDrawImage(imageContext, CGRectMake(0, 0, (CGFloat)width, (CGFloat)height), image);
	CGContextRelease(imageContext);
    
	// OpenGL用のテクスチャを生成
	glGenTextures(1, &texture);
	glBindTexture(GL_TEXTURE_2D, texture);
	glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, imageData);
    
	free(imageData);
	
	// 作成したテクスチャを返す
	return texture;
}
@end
